/*===============================================================
 * Project: libferit ( base netcode from Snarf 2.0 )
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : Ftp
 * Comment: Ftp module.
 *
 *-- History -----------------------------------------------
 *
 * 2000-02-14:
 * Mongoose - return server error code as int
 *
 * 2000-01-23:
 * Mongoose - Created
 *
 ==============================================================*/

#include "Ftp.h"
#include "Http.h"


Ftp::Ftp(int timeout) : TcpProtocol(timeout)
{
}


Ftp::~Ftp()
{
}


int Ftp::Interactive(Url *url, int ctrl_sock, int data_sock, int passive)
{
  char *line = NULL;
  char *cmd = NULL;
  int mode = 0;
  int retval;

  
  DisableClose();

  printf("Ftp::Interactive> %i %i %i\n", ctrl_sock, data_sock, passive);
  url->Print();

  for(;;)
  {
    cmd = Command();

    if (strcmp(cmd, "QUIT\r\n") == 0)
    {
      line = GetLine(url, ctrl_sock); /* 221 Goodbye */
  
      if (line)
	delete line;
  
      close(ctrl_sock);
      close(data_sock);
      return 0;
    }
    else if (strncmp(cmd, "LIST", 4) == 0)
    {
      mode = 2;
    }
    else if (strncmp(cmd, "NLST", 4) == 0)
    {
      mode = 2;
    }
    else if (strncmp(cmd, "LIST", 4) == 0)
    {
      mode = 2;
    }
    else
      mode = 0;

    /************
      SendControl(ctrl_sock, "SIZE ", url->Filename(), "\r\n", NULL);
      
      line = GetLine(url, ctrl_sock);
      
      if (line && CheckNumeric("213", line))
         url->OutfileSZ(atoi(line + 3)); 
      else 
         url->OutfileSZ(0);

      SendControl(ctrl_sock, "RETR ", url->Filename(), "\r\n", NULL);

     ***********/

    SendControl(ctrl_sock, cmd, NULL);
    line = GetLine(url, ctrl_sock);

    if (line)
      delete line;

    switch (mode)
    {
    case 1:
      retval = SaveStream(data_sock, url);
      break;
    case 2:
      if (data_sock == -1)
	data_sock = GetPassiveSock(url, ctrl_sock);

      if (data_sock == -1)
	data_sock = GetSock(url, ctrl_sock);

      retval = PrintStream(data_sock, url);

      line = GetLine(url, ctrl_sock); /* 226 Transfer complete */

      if (line)
	delete line;
      break;
    default:
      break;
  }
}
}


char *Ftp::Command()
{
  char *buffer = new char[64];
  char c;
  unsigned int i = 0;


  printf("\n\n >> ");
  fflush(stdout);

  while (i < 60)
  {
    fflush(stdin);
    c = fgetc(stdin);

    if (c == '\n')
    {
      buffer[i++] = '\r';
      buffer[i++] = '\n';
      buffer[i] = 0;
      i = 64;
    }
    else if (c == 8)
      i--;
    else
    {
      buffer[i++] = c;
      buffer[i] = 0;
    }
  }

  printf("\n");

  return buffer;
}


int Ftp::Download(Url *url, char *filename)
{
  char *line = NULL;
  int ctrl_sock = 0;
  int data_sock = 0;
  int passive = 1;
  int retval = 0;


   /* First of all, if this is proxied, just pass it off to the
      http module, since that's how we support proxying. */

   if (url->ProxyType() & FTP_PROXY) 
   {
      int n = 404;
      Http *http = new Http(10);
      n = http->Download(url, filename);
      delete http;

      return n;
   } 

   SetDefaults(url);

   if (filename)
     url->Outfile(filename);

   if (!url->Outfile())
      url->Outfile(url->Filename());  

   if (!(ctrl_sock = Connect(url->Hostname(), url->Port())))
      return -1;

   if (!(line = GetLine(url, ctrl_sock)))
      return -2;

   if (!CheckNumeric("220", line) ) 
   {
      if (line)
         delete line;

      printf("ERROR bad server greeting. \n");
      return -3;
   }

   SendControl(ctrl_sock, "USER ", url->Username(), "\r\n", NULL);

   if (!(line = GetLine(url, ctrl_sock)))
      return -4;

   /* Do the password dance */
   if (!CheckNumeric("230", line)) 
   {
      if(!CheckNumeric("331", line)) 
      {
         if (line)
            delete line;

         printf("ERROR: bad/unexpected response: %s", line);
         return 331;
      } 
      else 
      {
         if (line)
            delete line;

         SendControl(ctrl_sock, "PASS ", url->Password(), "\r\n", NULL);
         
         if (!((line = GetLine(url, ctrl_sock)) && CheckNumeric("230", line))) 
         {
            if (line)
	      delete line;

            printf("login failed\n");
            return 230;
         }

         if (line)
 	    delete line;
      }
   }
   
   /* Set binmode */
   SendControl(ctrl_sock, "TYPE I\r\n", NULL);

   if (!(line = GetLine(url, ctrl_sock)))
      return -5;

   if (line)
      delete line;
   
   if (url->Path()) 
   {
      SendControl(ctrl_sock, "CWD ", url->Path(), "\r\n", NULL);
      
      if (!((line = GetLine(url, ctrl_sock)) && CheckNumeric("250", line))) 
      {
         if (line)
	    delete line;

         Quit(ctrl_sock);
         return 250;
      }

      if (line)
         delete line;
   }
   
   /* Finally, the good stuff */

   /* Get a socket for reading. try passive first. */
   
   if (!(url->Options() & OPT_ACTIVE)) 
   {
      if ((data_sock = GetPassiveSock(url, ctrl_sock)) == -1)
         return -6;
   }

   if (!data_sock) 
   {
      if ((data_sock = GetSock(url, ctrl_sock)) < 1)
         return -7;
      else
         passive = 0;
   }

   /* Handle Resume */
   if (url->OutfileOFF() && (url->Options() & OPT_RESUME)) 
   {
      char numstring[BUFSIZE]; /* ugly hack */

      sprintf(numstring, "%ld", (long int )url->OutfileOFF());
      SendControl(ctrl_sock, "REST ", numstring, "\r\n", NULL);

      if (!((line = GetLine(url, ctrl_sock)) && CheckNumeric("350", line))) 
      {
         if (line)
	    delete line;

         Quit(ctrl_sock);

         printf("ERROR: server does not support FTP resume.\n");

         return 350;
      }

      if (line)
         delete line;
   }

   if (url->Filename()) 
   {
      SendControl(ctrl_sock, "SIZE ", url->Filename(), "\r\n", NULL);
      
      line = GetLine(url, ctrl_sock);
      
      if (line && CheckNumeric("213", line))
         url->OutfileSZ(atoi(line + 3)); 
      else 
         url->OutfileSZ(0);

      SendControl(ctrl_sock, "RETR ", url->Filename(), "\r\n", NULL);
   } 
   else
   {
     url->Outfile("file.lst");
     SendControl(ctrl_sock, "LIST\r\n", NULL); // NLST or LIST
   }

   if (!((line = GetLine(url, ctrl_sock)) &&
         (CheckNumeric("150", line) || CheckNumeric("125", line)))) 
   {
     /*================================================== 
      * Used to just drop session.
      * Now I do a LIST OR drop interactive mode. 
      *   - Mongoose
      ==================================================*/

     //if (line)
     //delete line;
     //
     //Quit(ctrl_sock);
     //return -8;

     url->Outfile("file.lst");
     SendControl(ctrl_sock, "LIST\r\n", NULL);

     //return Interactive(url, ctrl_sock, data_sock, passive);
   }

   if (!passive)
      data_sock = accept(data_sock, NULL, NULL);

   /*  url->outfile_size = guess_file_size(line); */

   if (line)
      delete line;

   if (strcmp(url->Outfile(), "file.lst") == 0)
   {
#define INTERACTIVE_FALLBACK

#ifdef INTERACTIVE_FALLBACK
     printf("\n\n=== File listing =============================\n");
     retval = PrintStream(data_sock, url);
     printf("\n==============================================\n\n");

     return Interactive(url, ctrl_sock, data_sock, passive);     
#elif
     printf("\n\n=== File listing =============================\n");
     retval = PrintStream(data_sock, url);
     printf("\n==============================================\n\n");
#endif
   }
   else
     retval = SaveStream(data_sock, url);

   if (retval < 0)
   {
     Quit(ctrl_sock);
     return -9;
   }

   line = GetLine(url, ctrl_sock); /* 226 Transfer complete */

   if (line)
      delete line;

   SendControl(ctrl_sock, "QUIT\r\n", NULL);

   line = GetLine(url, ctrl_sock); /* 221 Goodbye */

   if (line)
      delete line;

   close(ctrl_sock);
   close(data_sock);

   return retval;
}

void Ftp::Quit(int sock)
{
  if (sock > 0) 
  {
    write(sock, "QUIT\r\n", 6);
    close(sock);
  }
}

void Ftp::SetDefaults(Url *url)
{
   if (!url->Port())
      url->Port(21);

   if (!url->Username())
      url->Username("anonymous");

   if (!url->Password())
      url->Password("libferit@");
}

void Ftp::SendControl(int sock, char *string, ...)
{
   va_list args;
   char *line = NULL;
   char *newline;
   char *s = NULL;

   line = mon_strdup(string);
   
   va_start(args, string);
   s = va_arg(args, char *);

   while(s) 
   {
      newline = strconcat(line, s, NULL);

      if (line)
         delete line;

      line = newline;

      s = va_arg(args, char *);
   }

   va_end(args);

   printf("Sending request:\n%s\n", line);

   write(sock, line, strlen(line));

   if (line)
      delete line;
}

char *Ftp::GetLine(Url *url, int control)
{
   int bytes_read = 0;
   char *end;
   char buf[BUFSIZE+1];


   while((bytes_read = read(control, buf, BUFSIZE))) 
   {
      if (url->Options() & OPT_VERBOSE )
         fwrite(buf, 1, bytes_read, stderr);

      if ((buf[0] == '4' || buf[0] == '5') && 
          !((url->Options() & (OPT_VERBOSE | OPT_QUIET))) ) 
      {
         fwrite(buf, 1, bytes_read, stderr);
         return NULL;
      }

      /* In case there's a partial read */
      buf[bytes_read] = '\0';

      if (buf[bytes_read - 1] == '\n')
         buf[bytes_read - 1] = '\0';

      if (buf[bytes_read - 2] == '\r')
         buf[bytes_read - 2] = '\0';

      if (isdigit(buf[0]) && buf[3] == ' ') 
         return mon_strdup(buf);

      /* skip to last line of possibly multiple line input */

      if ((end = strrchr(buf, '\n')) ) 
      {
         end++;
         if (isdigit(end[0]) && end[3] == ' ')
            return mon_strdup(end);
      }
   }

   return NULL;
}


int Ftp::CheckNumeric(const char *numeric, const char *buf)
{
   return ((buf[0] == numeric[0]) && 
           (buf[1] == numeric[1]) &&
           (buf[2] == numeric[2]));
}

int Ftp::SockInit(struct sockaddr_in *sa, int control)
{
   socklen_t i;
   int sock;

   if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) 
   {
      perror("socket");
      return 0;
   }

   i = sizeof(*sa); 

   getsockname(control, (struct sockaddr *)sa, &i);
 
   sa->sin_port = 0 ; /* let system choose a port */

   if (bind (sock, (struct sockaddr *)sa, sizeof (*sa)) < 0) 
   {
      perror("bind");
      return 0;
   }

   return sock;
}

int Ftp::GetPassiveSock(Url *url, int control)
{
   unsigned char *addr;
   struct sockaddr_in sa;
   int sock;
   int x;
   char *line, *orig_line;


   SendControl(control, "PASV\r\n", NULL);

   if (!((line = GetLine(url, control)) && CheckNumeric("227", line))) 
   {
      if (line)
         delete line;

      return 0;
   }

   orig_line = line;

   if ( strlen(line) < 4 ) 
   {
      if (line)
         delete line;

      return 0;
   }

   if (!(sock = SockInit(&sa, control)))
      return -1;

   /* Skip the numeric response */
   line += 4;

   /* Then find the digits */
   
   while(!(isdigit(*line)))
      line++;

   /* Ugliness from snarf 1.x */

   sa.sin_family = AF_INET;
   addr = (unsigned char *)&sa.sin_addr;

   for(x = 0; x < 4; x++) 
   {
      addr[x] = atoi(line);
      line = strchr(line,',') + 1;
   }

   addr = (unsigned char *)&sa.sin_port ;
   addr[0] = atoi(line);
   line = strchr(line,',') + 1;
   addr[1] = atoi(line);
   
   if (connect(sock, (struct sockaddr *)&sa, sizeof(sa)) < 0 ) 
   {
      if (orig_line)
         delete orig_line;

      perror("connect");
      return -1;
   }

   if (orig_line)
      delete orig_line;

   return sock;
}

int Ftp::GetSock(Url *url, int control)
{
   struct sockaddr_in sa;
   unsigned char *addr;
   unsigned char *port;
   char *line;
   char port_string[BUFSIZE];
   unsigned int sock;
   socklen_t i;


   if (!(sock = SockInit(&sa, control)))
      return 0;

   if (listen(sock, 0) < 0 ) 
   {
      perror("listen");
      return 0;
   }
   
   i = sizeof(sa);

   getsockname(sock, (struct sockaddr *)&sa, &i);

   addr = (unsigned char *)(&sa.sin_addr.s_addr);
   port = (unsigned char *)(&sa.sin_port);

   sprintf(port_string, "PORT %d,%d,%d,%d,%d,%d\r\n", 
      addr[0], addr[1], addr[2], addr[3],
      port[0],(unsigned char)port[1]);

   SendControl(control, port_string, NULL);

   if (!((line = GetLine(url, control)) && CheckNumeric("200", line))) 
   {
      if (line)
         delete line;

      return 0;
   }

   if (line)
      delete line;
      
   return sock;
}

