/*===============================================================
 * Project: libferit ( base netcode from Snarf 2.0 )
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : Gopher
 * Comment: Gopher module.
 *
 *-- History -----------------------------------------------
 *
 * 2000-01-29:
 * Mongoose - Created
 *
 ==============================================================*/


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "Gopher.h"
#include "Url.h"
#include "utils.h"
#include "Http.h"


Gopher::Gopher(int timeout) : TcpProtocol(timeout)
{
}

Gopher::~Gopher()
{
}

int Gopher::Download(Url *url, char *filename)
{
   int sock = 0;
   char *request = NULL;
   

   if (filename)
      url->Outfile(filename);

   if (url->ProxyType() & GOPHER_PROXY) 
   {
      int n = 404;
      Http *http = new Http(10);
      n = http->Download(url, filename);
      delete http;

      return n;
   }

   if (!url->Port())
      url->Port(70);

   if (!url->Outfile())
      url->Outfile("gopher.txt");

   sock = Connect(url->Hostname(), url->Port());

   if (!sock)
      return -1;

   /* Mongoose: this is much better */
   request = NULL;

   if (!url->Path())
   {
     if (url->Filename())
     {
       request = new char[strlen(url->Filename()) + 8];
       sprintf(request, "%s\r\n", url->Filename());
     }
   }
   else if (url->Filename())
   {
     request = new char[strlen(url->Path()) + strlen(url->Filename()) + 8];
     sprintf(request, "%s%s\r\n", url->Path(), url->Filename());
   }

   if (!request)
      return -2;

   write(sock, request, strlen(request));

   /* Mongoose: free the memory, save the whales! */
   delete [] request;

   /* Never know the size of a gopher file */
   return SaveStream(sock, url);
}
