/*===============================================================
 * Project: libferit 
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : Url
 * Comment: Urls keep up with all the info needed to get files.
 *
 *-- History -----------------------------------------------
 *
 * 2000-01-23:
 * Mongoose - Created.
 ==============================================================*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "Url.h"
#include "utils.h"


Url::Url(char *url)
{
   _options     = OPT_NONE;
   _port        = 0;
   _protocol    = UNKNOWN_SESSION;
   _proxy_type  = NO_PROXY;
   _proxy       = NULL;
   _hostname    = NULL;
   _username    = NULL;
   _password    = NULL;
   _path        = NULL;
   _filename    = NULL;
   _outfile     = NULL;

   _outfile_size = 0;
   _outfile_offset = 0;

   UrlToSession(url);
}


Url::~Url()
{
   if (_proxy)
      delete _proxy;

   if (_hostname)
      delete _hostname;

   if (_password)
      delete [] _password;

   if (_username)
      delete [] _username;

   if (_outfile)
      delete [] _outfile;

   if (_path)
      delete [] _path;

   if (_filename)
      delete [] _filename;
}


void Url::PasswordPrompt()
{
  char buffer[64];
  char c;
  unsigned int i = 0;


  printf("\n\n\tPassword: ");
  fflush(stdout);

  while (i < 63)
  {
    fflush(stdin);
    //fscanf(stdin, "%c", &c);
    c = fgetc(stdin);

    if (c == '\n')
    {
      Password(buffer);
      i = 64;
    }
    else if (c == 8)
      i--;
    else
    {
      buffer[i++] = c;
      buffer[i] = 0;
    }
  }

  printf("\n\n");
}


/*-------------------------------------------------------------- 
 * Example url and strip steps:   
 *
 * ftp://mp3:mp3@ftp.cutie.com/Kristin_Hersh-San_Francisco.mp3 
 *
 *               0 - protocol strip
 *               1 - strip username:passwd, if exists
 *               2 - strip hostname
 *               3 - strip listing
 *               4 - finish buffering the rest as filename  
 *
 ***************************************************************
 * TODO:
 * 
 * Rewrite this to use tokens.
 *
 * <protocol>://<user>:<pass>@<host>/<path><file>   =>
 *
 * <S>         -> <protocol><auth><host><filename>
 * <protocol>  -> [A-Za-z]+://
 * <auth>      -> <user><pass>/ | /
 * <user>      -> [A-Za-z0-9]+:
 * <passwd>    -> [A-Za-z0-9]+@
 * <host>      -> <IP><port> | <hostname><port>
 * <IP>        -> <block>.<block>.<block>.<block>
 * <block>     -> [0-9][0-9][0-9]
 * <port>      -> :[0-9]+/ | /
 * <filename>  -> <path><file>
 * <path>      -> [A-Za-z0-9]+/<path> | NIL
 * <file>      ->  [A-Za-z0-9.-_]+
 *
 * -- Mongoose, 2001.10.14
  --------------------------------------------------------------*/
void Url::UrlToSession(char *url)
{
   char *buffer = new char[strlen(url)];
   char c = 'a';
   int i = 0, j = 0, stop = strlen(url);
   int len = 0, start_path = 0, start_filename = 0;
   bool port = false;


   if (!url || strcmp("", url) == 0)
   {
#ifdef DEBUG_URL
      printf("Url::UrlToSession> NULL url.\n");
#endif      

      _port = 0;
      delete buffer;
      return;      
   } 

   /*========== Strip protocol from url ==============*/

   if (!strncasecmp(url, "http://", 7))
   {
      _protocol = HTTP_SESSION;
      _port = 80;
      i += 7;
   }
   else if (!strncasecmp(url, "ftp://", 6))
   {
     _protocol = FTP_SESSION;
     _port = 21;
     i += 6;
   }
   else if (!strncasecmp(url, "gopher://", 9))
   {
      _protocol = GOPHER_SESSION;
      _port = 0;
      i += 9;
   }
   else if (!strncasecmp(url, "gnutella://", 11))
   {
      _protocol = GNUTELLA_SESSION;
      _port = 6346;
      i += 11;
   }
   else if (!strncasecmp(url, "smb://", 6))
   {
      printf("SMB is currently unsupported.\n");      
      _protocol = SMB_SESSION;
      i += 6;
      _port = 0;
      delete buffer;
      return;
   }
   else if (!strncasecmp(url, "nfs://", 6))
   {
      printf("NFS is currently unsupported.\n");      
      _protocol = NFS_SESSION;
      i += 6;
      _port = 0;
      delete buffer;
      return;
   }
   else if (!strncasecmp(url, "news://", 7))
   {
      printf("NNTP is currently unsupported.\n");      
      _protocol = NNTP_SESSION;
      _port = 119;
      i += 7;
      delete buffer;
      return;
   }
   else
   {
      printf("ERROR: Unknown session protocol.\n");
      printf("'%s' will not be added to queue/fetched.\n", url);
      _port = 0;
      delete buffer;
      return;
   }

   /*============= Strip username:password if exists ============*/

   if (strstr(url, "@"))
   {
      /*===== USERNAME ========*/

      j = 0;

      while((c = url[i++]) && (c != ':'))
      {
         if (c == '\0')
	 {
 	    printf("ERROR: '%s' may be invalid.\n", url);
            delete buffer;
            return;
	 }

         buffer[j++] = c;
      }

      buffer[j] = 0;

      _username = new char[strlen(buffer) + 1];
      strcpy(_username, buffer);

      /*======= PASSWORD ========*/

      j = 0;

      while((c = url[i++]) && (c != '@'))
      {
         if (c == '\0')
	 {
            printf("ERROR: %s may be invalid.\n", url);
            delete buffer;
            return;
	 }

         buffer[j++] = c;
      }

      buffer[j] = 0;

      _password = new char[strlen(buffer) + 1];
      strcpy(_password, buffer);
   }

   /*========= HOSTNAME ===========*/

   j = 0;

   while((c = url[i++]) && (c != '/'))
   {
      if (c == 0)
      {
         printf("ERROR: '%s' may be invalid.", url);
         delete buffer;
         return;
      }
      else if (c == ':')
      {
         port = true;
         break;
      }

      buffer[j++] = c;
   }

   buffer[j] = 0;

   _hostname = new char[strlen(buffer) + 1];
   strcpy(_hostname, buffer);

   /*======== PORT =============*/

   if (port)
   {
       j = 0;

      while((c = url[i++]) && (c != '/') && (c != '\0'))
      {
         if (c == '\0')
	 {
            printf("ERROR: %s may be invalid.", url);
            delete buffer;
            return;
	 }

         buffer[j++] = c;
      }

      buffer[j] = 0;

      _port = atoi(buffer);
   }

   /* ====== PATH ========= */

   if (i >= stop)   /* For proxy urls w/o paths and/or files */
   {
      _path = NULL;
      _filename = NULL;

      if (buffer)
         delete buffer;
 
      return;
   }

   j = 0;
   start_path = 0;

   while((c = url[i++]) && (c != '\0'))
   {
      buffer[j++] = c;

      if (c == '/')
         start_filename = j;
   }

   buffer[j] = 0;

   len = j;

   _path = new char[len + 2];

   strncpy(_path, buffer, start_filename);

   _path[start_filename] = 0;

#ifdef DEBUG
   printf("Url::UrlToSession> _path := '%s'\n", _path);
#endif

   if (strcmp(_path, "") == 0)
   {
     if (_path)
         delete _path;

     _path = new char[8];

      strcpy(_path, "/");
   }

   /* ====== FILENAME ========= */

   len = strlen(buffer) - start_filename;

   _filename = new char[len + 3];

   for (i = 0, j = start_filename; i < len; j++, i++)
      _filename[i] = buffer[j];    

   _filename[i] = 0;
   _filename[i+1] = 0;

#ifdef DEBUG
   printf("Url::UrlToSession> _filename := '%s'\n", _filename);
#endif

   if (buffer)
      delete buffer;
}

char *Url::FullUrl()
{
   unsigned int l = 0;
   char protocol[32];
   char *url;
   

   switch(_protocol)
   {
   case HTTP_SESSION:
     strcpy(protocol, "http");
     break;
   case FTP_SESSION:
     strcpy(protocol, "ftp");
     break;
   case GOPHER_SESSION:
     strcpy(protocol, "gopher");
     break;
   case GNUTELLA_SESSION:
     strcpy(protocol, "gnutella");
     break;
   default:
     printf("Url::FullUrl> Bad protocol\n");
     return NULL;
   }  

   if (!Hostname())
     return NULL;

   l = strlen(protocol) + strlen(Hostname()) + 32;

   if (Username() && Password())
   {
     l += strlen(Username()) + strlen(Password());

     if (Path() && Filename())
     {
       l += strlen(Filename()) + strlen(Path());

       url = new char[l];

       sprintf(url, "%s://%s:%s@%s:%i/%s%s", 
	       protocol, 
	       Username(), Password(), Hostname(), Port(), Path(), Filename());
     }
     else
     {
       url = new char[l];
       sprintf(url, "%s://%s:%s@%s:%i/", 
	       protocol, 
	       Username(), Password(),Hostname(), Port());
     }
   }
   else
   {
     if (Path() && Filename())
     {
       l += strlen(Filename()) + strlen(Path());
       url = new char[l];

       sprintf(url, "%s://%s:%i/%s%s", 
	       protocol, 
	       Hostname(), Port(), Path(), Filename());
     }
     else 
     {
       url = new char[l];
       sprintf(url, "%s://%s:%i/", 
	       protocol, 
	       Hostname(), Port());
     }
   }

#ifdef DEBUG
  printf("Url::FullUrl()> = %s\n", url);
#endif

  return url;
}


void Url::Print()
{
  Print("");
}


void Url::Print(char *s)
{ 
   printf("Url %s {\n  protocol: ", s);

   switch(_protocol)
   {
   case HTTP_SESSION:
     printf("http\n");
     break;
   case FTP_SESSION:
     printf("ftp\n");
     break;
   case GOPHER_SESSION:
     printf("gopher\n");
     break;
   case GNUTELLA_SESSION:
     printf("gnutella\n");
     break;
   default:
     printf("unknown\n");
     return;
   }

   printf("  host: %s\n", _hostname);
   printf("  port: %i\n", _port);
   printf("  user: %s\n", _username);
   printf("  passwd: %s\n", _password);
   printf("  path: %s\n", _path);
   printf("  filename: %s\n", _filename);
   printf("  outfile: %s\n}\n\n", _outfile);
}

void Url::OutfileOFF(off_t offset)
{
   _outfile_offset = offset;
}

off_t Url::OutfileOFF()
{
   return _outfile_offset;
}

void Url::OutfileSZ(off_t offset)
{
   _outfile_size = offset;
}

off_t Url::OutfileSZ()
{
   return _outfile_size;
}

char *Url::Outfile()
{
   return _outfile;
}

void Url::Outfile(char *s)
{
  if (!s)
    return;
 
  if (_outfile)
    delete _outfile; 

  _outfile = new char[strlen(s) + 1];
  strcpy(_outfile, s);
}

int Url::Options()
{
   return _options;
}

void Url::Options(int n)
{
   _options = n;
}

int Url::Protocol()
{
   return _protocol;
}


char *Url::Hostname()
{
   return _hostname;
}


void Url::Proxy(char *url, int proxy_type)
{
  if (!url || !url[0])
  {
    printf("Url::Proxy> ERROR: Null proxy url\n");
    return;
  }

  if (_proxy)
    delete _proxy;

  _proxy = new Url(url);
  _proxy_type = proxy_type;
}

Url *Url::Proxy()
{
   return _proxy;
}

int Url::ProxyType()
{
   return _proxy_type;
}

void Url::Port(int port)
{
   _port = port;
}

void Url::Username(char *user)
{
   if (_username)
      delete _username;

   _username = new char[strlen(user) + 1];
   strcpy(_username, user);
}

void Url::Password(char *passwd)
{
   if (_password)
      delete _password;

   _password = new char[strlen(passwd) + 1];
   strcpy(_password, passwd);
}

int Url::Port()
{
   return _port;
}

char *Url::Username()
{
   return _username;
}

char *Url::Password()
{
   return _password;
}

char *Url::Path()
{
   return _path;
}

char *Url::Filename()
{
   return _filename;
}

void Url::Filename(char *s)
{
   if (_filename)
     delete _filename;

   _filename = new char[strlen(s) + 1];
   strcpy(_filename, s);  
}

#ifdef __TEST

int main(int argc, char **argv)
{
   Url *url;


   if (argc == 2)
   {
      url = new Url(argv[1]);
      url->Print();
      delete url;
   }

   return 1;
}

#endif
