/*===============================================================
 * Project: libferit
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : Url
 * Comment: Urls keep up with all the info needed to get files.
 *
 *-- History -----------------------------------------------
 *
 * See source file.
 *
 ==============================================================*/

#ifndef __MONGOOSE_LIBFERIT_URL_H
#define __MONGOOSE_LIBFERIT_URL_H

#include <sys/types.h>


enum url_services                   /* Service types */
{
   UNKNOWN_SESSION   =  0,
   HTTP_SESSION      =  1,
   FTP_SESSION       =  2,
   GOPHER_SESSION    =  4,
   SMB_SESSION       =  8,
   NFS_SESSION       =  16,
   FINGER_SESSION    =  32,
   NNTP_SESSION      =  64,
   GNUTELLA_SESSION  = 128
};

enum url_options                    /* Options */
{
   OPT_NONE          =  0,
   OPT_RESUME        =  1,
   OPT_NORESUME      =  2,
   OPT_VERBOSE       =  4,
   OPT_QUIET         =  8,
   OPT_ACTIVE        =  16,
   OPT_BE_MSIE       =  32,
   OPT_BE_MOZILLA    =  64,
   OPT_BE_SNARF      =  128,
   OPT_PROGRESS      =  256,
   OPT_INTERACTIVE   = 1024
};

enum url_proxy                      /* Proxies */
{
   NO_PROXY          =  0,
   GOPHER_PROXY      =  1,
   FTP_PROXY         =  2,
   HTTP_PROXY        =  4,
   NNTP_PROXY        =  8
};

class Url
{
private:

   int  _options;
   int  _port;
   int  _protocol;
   int  _proxy_type;
   Url  *_proxy;
   char *_hostname;
   char *_username;
   char *_password;
   char *_path;
   char *_filename;
   char *_outfile;
   off_t _outfile_size;
   off_t _outfile_offset;


public:

   Url(char *url);
   /*------------------------------------------------------
    * Pre  : Url is a valid url, OR the will be set 
    *        be set by a valid call to UrlToSession() later 
    *
    * Post : Url is constructed
    *------------------------------------------------------
    * 20000123: 
    * Mongoose - Created
    ------------------------------------------------------*/

   ~Url();
   /*------------------------------------------------------
    * Pre  : Url is allocated
    * Post : url is deconstructed
    *------------------------------------------------------
    * 20000123: 
    * Mongoose - Created
    ------------------------------------------------------*/

   void PasswordPrompt();
   /*------------------------------------------------------
    * Pre  : This Url has username
    *        Password entered must be less than 64 bytes
    *
    * Post : Resets or sets password for this Url
    *        by interactive prompt
    *------------------------------------------------------
    * 20010404: 
    * Mongoose - Created
    ------------------------------------------------------*/

   void UrlToSession(char *url);
   /*------------------------------------------------------
    * Pre  : url is valid URL string
    * Post : url is parsed into sections
    *------------------------------------------------------
    * 20010402:  
    * Mongoose - Fixed some comments and string code
    *
    * 20000123: 
    * Mongoose - Created
    ------------------------------------------------------*/

   char *FullUrl();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : 
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void Print();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void Print(char *s);
   /*------------------------------------------------------
    * Pre  : s is a valid string
    * Post : Url is printed to stdout with s as identifer
    *
    *-- History ------------------------------------------
    *
    * 2001.10.14: 
    * Mongoose - Created
    ------------------------------------------------------*/

   off_t OutfileOFF();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void OutfileOFF(off_t offset);
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   off_t OutfileSZ();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void OutfileSZ(off_t offset);
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : 
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   char *Outfile();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : 
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void Outfile(char *s);
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   int Options();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : 
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void Options(int n);
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   int Protocol();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : 
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   char *Hostname();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void Proxy(char *url, int proxy_type);
   /*------------------------------------------------------
    * Pre  : Pass a proxy url (string) and proxy type
    * Post : Parses into usable form for class
    *------------------------------------------------------
    * 20010404:  
    * Mongoose - Handling of multiple calls and errors
    *
    * 20000123: 
    * Mongoose - Created
    ------------------------------------------------------*/

   Url *Proxy();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   int ProxyType();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void Port(int port);
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void Username(char *user);
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void Password(char *passwd);
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   int Port();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   char *Username();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   char *Password();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   char *Path();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   char *Filename();
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/

   void Filename(char *s);
   /*-------------------------------------------
    * Created  : 2000 01 23 by Terry Hendrix II
    * Modified : NA
    * Pre      : 
    * Post     : 
    -------------------------------------------*/
};

#endif




	
