/*===============================================================
 * Project: libferit
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : NA
 * Comment: This is the client ferit.
 * 
 *-- History -----------------------------------------------
 *
 * 1999-08-18:
 * Mongoose - Created, based on prototype done in ANSI C
 ==============================================================*/
#include "Url.h"
#include "Ftp.h"
#include "Http.h"
#include "Gopher.h"
#include "Gnutella.h"

#ifdef MODULES
#include <dlfcn.h>
#endif

#include "utils.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

char *PROXY_URL = NULL;


void version()
{
    printf("Ferit %s\n", CLIENT_VERSION);
    printf("by Terry 'Mongoose' Hendrix <mongoose@icculus.org>\n");  
    printf("http://www.icculus.org/~mongoose/projects/\n\n");  
}

void help()
{
    printf("ferit [COMMAND1] [COMMAND2]...\n\n");

    printf("Ferit commands:\n");
    printf("   URL [FILENAME]  \tFetch using just using the URL.\n");
    printf("   -               \tDump data to stdout.\n");
    printf("   --batch | b FILE\tDownload list of urls from FILE.\n");
    printf("   --clean         \tSave as a clean unix filename.\n");
    printf("   --dir DIR       \tSet download directory.\n");
    printf("   --modtest       \tTest for module support in the build.\n");
    printf("   --mozilla | m   \tSpoof mozilla user agent string.\n");
    printf("   --msie          \tSpoof msie user agent string.\n");
    printf("   --no-rc         \tDon't use rc file.\n");
    printf("   --pause         \tWait for input before exit.\n");
    printf("   --proxy URL     \tUse this proxy.\n");
    printf("   --quiet | q     \tSurpress progress bar.\n");
    printf("   --resume        \tResume broken downloads.\n");
    printf("   --retry NUM     \tFailed fetch rety NUM times. ( -1 inf )\n");
    printf("   --timeout NUM   \tSet TCP timeout to NUM seconds.\n");
    printf("   --trytime NUM   \tSet retry pause to NUM seconds.\n");
    printf("   --verbose | v   \tGet more session information.\n");
    printf("\n");
}


void test_module()
{
#ifdef USE_DL_MODULE_SUPPORT
   void *handle;
   void *test_f;
   char *error;


   printf("[Module test]\n");

   if (!(handle = dlopen ("./test.so", RTLD_NOW))) //RTLD_LAZY)))
   {
      printf("\tERROR: Couldn't fine module 'test.so'.\n");
   }
   else
   {
      printf("\tTest module opened.\n");

      double (*test_f)(double) = (double (*)(double))dlsym(handle, "test");

      if ((error = dlerror()) != NULL)  
      {
          fprintf (stderr, "\tERROR: %s\n", error);
          exit(1);
      }

      printf("Main> Module returned: %f\n", (*test_f)(128.7));

      dlclose(handle);
   }
#else
   printf("This build is not module enabled.\n");
#endif
}


int fetch(Url *url, int timeout)
{
  int err = 0;
  TcpProtocol *client = NULL;
  

  if (!url)
    return -10000; 
 
  if (url->Options() & OPT_VERBOSE)
    url->Print("fetch()");

  if (url->Password() && strcmp(url->Password(), "ASK") == 0)
    url->PasswordPrompt();
  
  switch (url->Protocol())
  {
  case FTP_SESSION:
    client = new Ftp(timeout);
    break;
  case HTTP_SESSION:
    client = new Http(timeout);
    break;
  case GOPHER_SESSION:
    client = new Gopher(timeout);
    break;
  case GNUTELLA_SESSION:
    client = new Gnutella(timeout);
    break;
  default:
    printf("Sorry the protocol used in the URL is unknown.\n");
    
    if (client)
    {
      delete client;
    }

    return -20000;
  }
  
  if (client)
  {
    err = client->Download(url, NULL);

    if (!(url->Options() & OPT_QUIET))
      printf("Error status: %i\n", err);

    delete client;
  }

  return err;
}


void read_rc(char *s)
{
   FILE *f = fopen(s, "r");
   char c;
   char buffer[256];
   char name[64];
   int i = 0;


   if (!f)
   {
     printf("File \'%s\' not found.\n", s);
   }
   else
   {
      while(fscanf(f, "%c", &c) != EOF)
      {
         switch(c)
	 {
	 case '\n':
            if (buffer[0] == '#')
	    {
               i = 0;
               continue;
	    }
         case ' ':
         case '\t':
            break;
         case '=':
 	    buffer[i] = 0;
            strcpy(name, buffer);
            i = 0;
	    break;
         case ';':
	    buffer[i] = 0;

	    if (strcmp(name, "PROXY_URL") == 0)
	    {
	      PROXY_URL = new char[i + 2];
	      strcpy(PROXY_URL, buffer);
	    }

	    i = 0;
	    break;
         default:
            buffer[i++] = c;
	 }
      }

      fclose(f);
   }
}


void setup()
{
   char *s;


   s = new char[strlen(getenv("HOME")) + strlen("/.ferit") + 2];
   sprintf(s, "%s%s", getenv("HOME"), "/.ferit");

   read_rc(s);
}


int script(char *filename, Url *url, int timeout)
{
  char buffer[1024];
  char c;
  FILE *f;
  int len = 0;
  int err = 0;


  f = fopen(filename, "r");

  if (!f)
  {
    perror("script> ERROR: ");
    return -1;
  }

  while (fscanf(f, "%c", &c) != EOF)
  {
    if (len > 1022)
      len--;

    switch (c)
    {
    case '\n':
      buffer[len] = 0;
      
      if (buffer[0] != '#' && buffer[0] != ' ' && buffer[0] != 0)
      {
	url->UrlToSession(buffer);
	url->Outfile(url->Filename());
	err += fetch(url, timeout);
      }

      len = 0;  
      
      break;
    default:
      buffer[len++] = c;
    }
  }

  fclose(f);
  return err;
}


int isProtocol(char *url)
{
	int known = 0;

	if (strncmp(url, "http://", strlen("http://")) == 0)
	{
		known = 1;
	}
	else if (strncmp(url, "ftp://", strlen("ftp://")) == 0)
	{
		known = 2;
	}
	else if (strncmp(url, "gnutella://", strlen("gnutella://")) == 0)
	{
		known = 3;
	}
	else if (strncmp(url, "gopher://", strlen("gopher://")) == 0)
	{
		known = 4;
	}

	return known;
}


int main(int argc, char *argv[])
{
  Url *url = NULL;
  char *script_file = NULL;
  bool clean = false;
  bool pause = false;
  bool use_rc = true;
  int mode = 0;
  int i = 0;
  int timeout = 10;
  int trytime = 10;
  int retry = 1;
  

  url = new Url("");
  url->Options(OPT_RESUME | OPT_PROGRESS);

  if (PROXY_URL)
  {
    url->Proxy(PROXY_URL, HTTP_PROXY + FTP_PROXY + GOPHER_PROXY);
  }
 
  for (i = argc; i > 1; i--)
  {
    if (strcmp(argv[i - 1], "--version") == 0)
    {
      version();
      exit(0);
    }
    else if ((strcmp(argv[i - 1], "-") == 0))
    {
      url->Options(OPT_RESUME | OPT_QUIET);
      url->Outfile("/dev/stdout");
    }
    else if ((strcmp(argv[i - 1], "--verbose") == 0) ||
	     (strcmp(argv[i - 1], "v") == 0))
      url->Options(url->Options() | OPT_VERBOSE);
    else if (strcmp(argv[i - 1], "--mozilla") == 0 ||
	     (strcmp(argv[i - 1], "m") == 0))
      url->Options(url->Options() | OPT_BE_MOZILLA);
    else if (strcmp(argv[i - 1], "--msie") == 0)
      url->Options(url->Options() | OPT_BE_MSIE);
    else if (strcmp(argv[i - 1], "--resume") == 0)
      url->Options(url->Options() | OPT_RESUME);
    else if (strcmp(argv[i - 1], "--quiet") == 0 ||
	     (strcmp(argv[i - 1], "q") == 0))
      url->Options(url->Options() | OPT_QUIET);
    else if (strcmp(argv[i - 1], "--clean") == 0)
      clean = true;
    else if (strcmp(argv[i - 1], "--pause") == 0)
      pause = true;
    else if (strcmp(argv[i - 1], "--no-rc") == 0)
      use_rc = false;
    else if (strcmp(argv[i - 1], "--help") == 0)
    {
      help();
      exit(0);
    }
    else if (strcmp(argv[i - 1], "--modtest") == 0)
    {
      test_module();
      exit(0);
    }
    else if (isProtocol(argv[i - 1]))
    {
      url->UrlToSession(argv[i - 1]);
      mode = 1;
    }
    else if (strcmp(argv[i - 1], "--retry") == 0)
      ;
    else if (strcmp(argv[i - 1], "--trytime") == 0)
      ;
    else if (strcmp(argv[i - 1], "--timeout") == 0)
      ;
    else if ((strcmp(argv[i - 1], "--batch") == 0) ||
	     (strcmp(argv[i - 1], "b") == 0))
      ;
    else if (strcmp(argv[i - 1], "--fetch") == 0)
      ;
    else if (strcmp(argv[i - 1], "--proxy") == 0)
      ;
    else if (strcmp(argv[i - 1], "--dir") == 0)
      ;
    else
    {
      if (argc > 2 && strcmp(argv[i - 2], "--proxy") == 0)
      {
	if (argv[i - 1])
	  url->Proxy(argv[i - 1], HTTP_PROXY + FTP_PROXY + GOPHER_PROXY);
	
	i--;
	continue;
      }
      else if (argc > 2 && strcmp(argv[i - 2], "--dir") == 0)
      {
	if (argv[i - 1])
	{
		char filename[256];
		bool ending = false;
		unsigned int len = strlen(argv[i - 1]);
		

		if (argv[i - 1][len-1] == '/')
			ending = true;

		snprintf(filename, 256, "%s%s%s",
			 argv[i - 1],  ending ? "" : "/", url->Filename());
		url->Outfile(filename);
	}
	
	i--;
	continue;
      }
      else if (argc > 2 && strcmp(argv[i - 2], "--retry") == 0)
      {
	if (argv[i - 1])
	{
	  retry = atoi(argv[i - 1]);
	}
	
	i--;
	continue;
      }
      else if (argc > 2 && strcmp(argv[i - 2], "--timeout") == 0)
      {
	if (argv[i - 1])
	{
	  timeout = atoi(argv[i - 1]);
	}
	
	i--;
	continue;
      }
      else if (argc > 2 && strcmp(argv[i - 2], "--trytime") == 0)
      {
	if (argv[i - 1])
	{
	  trytime = atoi(argv[i - 1]);
	}
	
	i--;
	continue;
      }
      else if (argc > 2 && (strcmp(argv[i - 2], "--batch") == 0 ||
	     (strcmp(argv[i - 2], "b") == 0)))
      {
	if (argv[i - 1])
	{
	  if (!script_file)
	  {  
	    script_file = new char[strlen(argv[i - 1]) + 1];
	    strcpy(script_file, argv[i - 1]);
	    mode = 2;
	  }
	}
              
	i--;
	continue;
      }
      else if (argc > 2 && isProtocol(argv[i - 2]) &&
	       argv[i - 1][0] != '-')
      {
	if (argv[i - 1])
	  url->Outfile(argv[i - 1]);
	
	i--;
	
	if (argv[i - 1])
	{
	  url->UrlToSession(argv[i - 1]);
	  mode = 1;
	}
              
	i--;
	continue;
      }
      
      help();
      printf("Unknown command: %s\n", argv[i - 1]);
      exit(0);
    }
  }

  if (use_rc)
    setup();

  if (clean)
    url->Outfile(clean_filename(url->Outfile()));

  if (url->Options() & OPT_VERBOSE)
  {
    //version();
    printf("\tPROXY_URL = \'%s\';\n\n", PROXY_URL);
  }

  //  url->Print();
  //  printf("\n");

  switch (mode)
  {
  case 1:    
    for (i = 0; i < retry; i++)
    {
      if (!fetch(url, timeout))
	break;
      
      if ((i+1) < retry)
      {
	printf("\n----------------------------------\nAttempt %i/%i\n\n",
	       i+1, retry);
	sleep(trytime);
      }
    }
    break;
  case 2:
    for (i = 0; i < retry; i++)
    {
      if (!script(script_file, url, timeout))
	break;
      
      if ((i+1) < retry)
      {
	printf("\n----------------------------------\nAttempt %i/%i\n\n",
	       i+1, retry);
	sleep(trytime);
      }
    }
    break;
  default:
    help();
  }

  if (url)
    delete url;

  if (pause)
  {
    char c;
    
    printf("\n[Press <Enter> to exit]\n");
    scanf("%c", &c);
  }
  
  return 0;
}






