/*===============================================================
 * Project: libsnarf ( base netcode from Snarf 2.0 )
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : Nntp
 * Comment: Nntp module.
 *
 *-- History -----------------------------------------------
 *
 * 2000-02-13:
 * Mongoose - Created ( based on Rael's Binary Grabber v1.2 )
 *
 ==============================================================*/

#include "Nntp.h"
#include "Http.h"

void *module_load_nntp(){return (new Nntp());}

void nntp_fetch(void *object, char *f)
{
   Nntp *nntp = (Nntp *)object;

   nntp->Fetch(NULL, f);
}

Nntp::Nntp() : Common()
{
}

Nntp::~Nntp()
{
}

int Nntp::Fetch(Url *url, char *filename)
{
   FILE *out		= NULL;
   char *line		= NULL;
   int 	sock		= 0;
   int 	data_sock	= 0;
   int	retval		= 0;


   /* First of all, if this is proxied, just pass it off to the
      http module, since that's how we support proxying for now. */

   if (url->ProxyType() & NNTP_PROXY) 
   {
      int n = 404;
      Http *http = new Http();
      n = http->Fetch(url, filename);
      delete http;

      return n;
   } 

   printf("Nntp module loaded.\nNntp support is in devel.\n");
   return -1;

   SetDefaults(url);

   if (filename)
     url->Outfile(filename);

   if (!url->Outfile())
      url->Outfile(url->Filename());  

   if (!(sock = TCPConnect(url->Hostname(), url->Port())))
      return 0;

   if (!(line = GetLine(url, sock)))
      return 0;

   if (!CheckNumeric("220", line) ) 
   {
      if (line)
         delete line;

      printf("ERROR bad server greeting. \n");
      return 0;
   }

   SendControl(sock, "AUTHINFO USER ", url->Username(), "\r\n", NULL);

   if (!(line = GetLine(url, sock)))
      return 0;

   /* FIXME: 230 and 331 are FTP controls */
   if (!CheckNumeric("230", line)) 
   {
      if(!CheckNumeric("331", line)) 
      {
         if (line)
            delete line;

         printf("ERROR: bad/unexpected response: %s", line);
         return 0;
      } 
      else 
      {
         if (line)
            delete line;

         SendControl(sock, "AUTHINFO PASS ", url->Password(), "\r\n", NULL);
         
         if (!((line = GetLine(url, sock)) && CheckNumeric("502", line))) 
         {
            if (line)
	      delete line;

            printf("nntp login failed\n");
            return 502;
         }

         if (line)
 	    delete line;
      }
   }


   
   fclose(out);
   close(sock);
   close(data_sock);

   return retval;
}

#ifdef NNTP_DEVELOPMENT
int Nntp::ProcessHeader(Url *url, int sock) 
{
   char *tline;
   char tmpstring[80];
   char *msgID;

   SendControl(sock, "HEAD\n", NULL);


   tline = GetLine(url, sock);

   msgID = strtok(tline," "); 
   msgID = strdup(strtok(NULL," "));
   output_msgID(msgID);

   /* Loop until the first blank line.
      First line after blank is first message text line. */

   do 
   {
      if (tline) 
         delete [] tline;

      tline = GetLine(url, sock);

      if (strncmp(tline, "Subject:", 8) == 0) 
      {
         myMessage = new message(tline);

         if (!(*isSpam)) 
            output_subject((char *)(tline+9));
      } 
      else if ((_spam_threshold != -1) && (strncmp(tline, "Newsgroups:", 11))) 
      {
         int tmp1 = 1;

         for (int w = 0; w < strlen(tline); w++) 
            if (tline[w] == ',') tmp1++;   

         if (tmp1 > spamThreshold) 
         {
            *isSpam = true;

            sprintf(tmpstring, 
                    "Skipping spam message (%d cross-posts)...",
                    tmp1);

            output_subject(tmpstring);
         }
      }
   } while (!(tline[0]=='.'));

   delete [] tline;
}

#endif

void Nntp::Quit(int sock)
{
   if (sock) 
   {
      write(sock, "QUIT\r\n", 6);
      close(sock);
   }
}

void Nntp::SetDefaults(Url *url)
{
   if (!url->Port())
      url->Port(115);

   if (!url->Username())
      url->Username("anonymous");

   if (!url->Password())
      url->Password("libsnarf@");
}

void Nntp::SendControl(int sock, char *string, ...)
{
   va_list args;
   char *line = NULL;
   char *newline;
   char *s = NULL;

   line = mon_strdup(string);
   
   va_start(args, string);
   s = va_arg(args, char *);

   while(s) 
   {
      newline = strconcat(line, s, NULL);

      if (line)
         delete line;

      line = newline;

      s = va_arg(args, char *);
   }

   va_end(args);

   write(sock, line, strlen(line));

   if (line)
      delete line;
}

char *Nntp::GetLine(Url *url, int control)
{
   int bytes_read = 0;
   char *end;
   char buf[BUFSIZE+1];


   while((bytes_read = read(control, buf, BUFSIZE))) 
   {
      if (url->Options() & OPT_VERBOSE )
         fwrite(buf, 1, bytes_read, stderr);

      if ((buf[0] == '4' || buf[0] == '5') && 
          !((url->Options() & (OPT_VERBOSE | OPT_QUIET))) ) 
      {
         fwrite(buf, 1, bytes_read, stderr);
         return NULL;
      }

      /* In case there's a partial read */
      buf[bytes_read] = '\0';

      if (buf[bytes_read - 1] == '\n')
         buf[bytes_read - 1] = '\0';

      if (buf[bytes_read - 2] == '\r')
         buf[bytes_read - 2] = '\0';

      if (isdigit(buf[0]) && buf[3] == ' ') 
         return mon_strdup(buf);

      /* skip to last line of possibly multiple line input */

      if ((end = strrchr(buf, '\n')) ) 
      {
         end++;
         if (isdigit(end[0]) && end[3] == ' ')
            return mon_strdup(end);
      }
   }

   return NULL;
}


int Nntp::CheckNumeric(const char *numeric, const char *buf)
{
   return ((buf[0] == numeric[0]) && 
           (buf[1] == numeric[1]) &&
           (buf[2] == numeric[2]));
}

int Nntp::SockInit(struct sockaddr_in *sa, int control)
{
   socklen_t i;
   int sock;

   if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) 
   {
      perror("socket");
      return 0;
   }

   i = sizeof(*sa); 

   getsockname(control, (struct sockaddr *)sa, &i);
 
   sa->sin_port = 0 ; /* let system choose a port */

   if (bind (sock, (struct sockaddr *)sa, sizeof (*sa)) < 0) 
   {
      perror("bind");
      return 0;
   }

   return sock;
}

int Nntp::GetSock(Url *url, int control)
{
   struct sockaddr_in sa;
   unsigned char *addr;
   unsigned char *port;
   char *line;
   char port_string[BUFSIZE];
   unsigned int sock;
   socklen_t i;


   if (!(sock = SockInit(&sa, control)))
      return 0;

   if (listen(sock, 0) < 0 ) 
   {
      perror("listen");
      return 0;
   }
   
   i = sizeof(sa);

   getsockname(sock, (struct sockaddr *)&sa, &i);

   addr = (unsigned char *)(&sa.sin_addr.s_addr);
   port = (unsigned char *)(&sa.sin_port);

   sprintf(port_string, "PORT %d,%d,%d,%d,%d,%d\r\n", 
      addr[0], addr[1], addr[2], addr[3],
      port[0],(unsigned char)port[1]);

   SendControl(control, port_string, NULL);

   if (!((line = GetLine(url, control)) && CheckNumeric("200", line))) 
   {
      if (line)
         delete line;

      return 0;
   }

   if (line)
      delete line;
      
   return sock;
}

