/*===============================================================
 * Project: libferit
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : Spider
 * Comment: This is the spider module of libferit.
 * 
 *-- History -----------------------------------------------
 *
 * 2000-02-26:
 * Mongoose - Decided to strip out all bloat from goosehunt
 *             removed host indexing, cache statisics, etc
 *
 *            Moved all code into this module to make a
 *             prototype Goosehunt LITE module for libferit
 *
 *
 * 1999-08-18:
 * Mongoose - Created, based on prototype done in ANSI C
 ==============================================================*/

#include "Spider.h"

#include "Ftp.h"
#include "Http.h"
#include "Gopher.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>


int SLEEP_TIME   = 3;
int MAX_PAGES    = 5;
int MAX_BRANCHES = 5;


char *VERSION = "Goosehunt 0.0.20-LITE";
char *CACHE_PATH;
char *DATA_PATH;
char *PROXY_URL;

void append(char *filename, char *s);
int ripper(FILE *fi, FILE *f, int type);
void parse_links(char *infile, char *outfile, int type);
char gcompare(char *s0, char *s1);


void shutdown_goosehunt()
{
   printf("Send comments, bug reports, and code to:\n\n");
   printf("     Email: <Mongoose>stu7440@westga.edu\n");
   printf("     Subject: %s\n\n", VERSION);
   exit(0);
}

//strncasecmp();

char gcompare(char *s0, char *s1)
{
   char buffer[32];


   strncpy(buffer, s1, 30);
   buffer[strlen(s0)] = 0;

   return (caseless_strcmp(buffer, s0));
}

int ripper(FILE *fi, FILE *f, int type)
{
   char c; 
   int save = 0, done = 0, quote = 0, append = 0;
   char buffer[256];
   char url[16];
   int i, size = 0;


   while(!done && fscanf(fi, "%c", &c) != EOF)
   {
      if (size > 255)
         return c;

      if (append && c != '"')
      {
         buffer[size++] = c;
      }

      if (c == '"')
      {
         switch(quote)
         {
         case 0:
            append = 1;
            quote = 1;
            break;
         case 1:
            append = 0;
            done = 1;
            break;
         }
      }
   }

   buffer[size] = 0;

   /* Urls this small worth anything? */
   if (size < 10)
      return 0;
  
   for (i = 0; i < 16; i++)
      url[i] = buffer[i];

   url[15] = 0;


   /* Print 3 char file ext */
   // printf("%c%c%c\n", buffer[size-3],buffer[size-2],buffer[size-1]);

   if (type == 1) 
   {

      if (((buffer[size - 3] == 'M') || (buffer[size - 3] == 'm')) && 
          ((buffer[size - 2] == 'P') || (buffer[size - 2] == 'p')) && 
          (buffer[size - 1] == '3'))
      {
         printf(".");
         save = 1;
      }
   }
   else if (gcompare("/cgi-bin", url) == 0)
   {
      printf("|");
      save = 1;      
      fprintf(f, "http://mp3search.lycos.com"); // base_url);
   }

   if (save)
   {
      for (i = 0; i < size; i++)
         fprintf(f, "%c", buffer[i]);

      fprintf(f, "\n");
   }

   return c;
}

/* Mongoose: Not to be used in threaded code
             Very slow
             Very much a temp hack */
void append(char *filename, char *s)
{
   FILE *f;   
   char buffer[512];
   char c;
   int found = 0, i = 0;


   if (!s || strcmp(s, "") == 0)
      return;

   f = fopen(filename, "a+");

   fseek(f, 0L, SEEK_SET);

   while(fscanf(f, "%c", &c) != EOF && !found)
   {
      if (i > 510)
	 buffer[511] = c;
      else
         buffer[i++] = c;
      
      if (c == '\n')
      {
 	 buffer[i] = 0;
         i = 0;

         if (strcmp(buffer, s) == 0)
            found = 1;
      }
   }

   fseek(f, 0L, SEEK_END);

   if (!found)
      fprintf(f, "%s", s);

   fclose(f);
}

void parse_links(char *infile, char *outfile, int type)
{
   FILE *inf;
   FILE *outf;
   char c, rip_link = 0;
   char buffer[32];
   unsigned int i = 0; 
   int links = 0;


   inf = fopen(infile, "r");

   outf = fopen(outfile, "a");

   if (!inf)
   {
      perror(infile);
      return;
   }


   while(fscanf(inf, "%c", &c) != EOF)
   {
      if (rip_link == 0)
      {
         if (i > 10)
            i = 0;

         buffer[i++] = c;  
         buffer[i] = 0;
      }

      if (rip_link != 1 && gcompare(buffer, "A HREF") == 0)
      {

         if (i == strlen("A HREF"))
	 {
            rip_link = 1;
            links++;

            if (ripper(inf, outf, type) == EOF)
               break;
	 }         
      }
      else
      {
	 rip_link = 0;
         i = 0;
      }
   }

   fclose(inf);
   fclose(outf);

   printf("\nRipped %i raw links.\n", links);
}

Spider::Spider()
{
   _tagged = false;
}

Spider::~Spider()
{
}

void Spider::QueryNextPage(char *lnkfile, char *name, int page, char *key)
{
   char tmpfile[64]; 
   char urlfile[33]; 
   char buffer[256];
   FILE *f;
   char c;
   int i = 0, len = 254, branch = 0, err = 0;

 
   if (SLEEP_TIME > 0)
   {
      printf("Sleeping %i seconds.\n", SLEEP_TIME);
      sleep(SLEEP_TIME);
   }

   sprintf(tmpfile, "%s%i-%i.tmp", name, page, branch++);
   sprintf(urlfile, "%s%i.urls", name, page);

   f = fopen(lnkfile, "r");

   if (!f)
   {
      printf("No link file generated...\n");
      return;
   }

   while(fscanf(f, "%c", &c) != EOF)
   {
      if (i > len)
        return;

      if (page >= MAX_PAGES || branch >= MAX_BRANCHES)
        continue;

      if (c == '\n') /* Found a url in link file */
      {
         buffer[i++] = 0;

         /* Get another cache file to get new lnk files */
         sprintf(tmpfile, "%s%i-%i.tmp", name, page, branch++);

         /* Get the html page the url points to */
         err = Get(buffer, tmpfile);
         //Handle(err, url, filename);

         /* Rip out all the ftp|http links to a url file */
         parse_links(tmpfile, urlfile, 1); /* Change 1 to ext later */

         /* Append these urls into one spider url file for key */
         Cache(urlfile, key);

         /* Try to get links off this new page */
         page++;
         QueryPage(tmpfile, name, page, key);

         /* Clean up cache */
         unlink(tmpfile);
         unlink(urlfile);

         i = 0;         
      }
      else
      {
         buffer[i++] = c;
         buffer[i] = 0;
      }
   }

   fclose(f);
}

int Spider::Cache(char *tmp, char *key)
{
   FILE *fout = fopen(key, "a+");
   FILE *fin = fopen(tmp, "r");
   char c;
   int i;
   char buffer[256];

   if (!fout)
   {
      printf("Error opening url cache.\n");
      return 0;
   }

   if (!fin)
   {
      printf("Error opening tmp url cache.\n");
      return 0;
   }

   if (!_tagged)
   {
      fprintf(fout, "# Generated by Ferit\n");  
      fprintf(fout, "# (C) 1999-2000 Terry 'Mongoose' Hendrix II");
      fprintf(fout, " <stu7440@westga.edu>\n");
      fprintf(fout, "# http://www.westga.edu/~stu7440/projects\n");  
      _tagged = true;
   }

   fclose(fout);

   i = 0;

   while(fscanf(fin, "%c", &c) != EOF)
   {
      if (i > 255)
      {
	 buffer[0] = '#';
         i = 254;
      }

      buffer[i++] = c;
      
      if (c == '\n')
      {
 	 buffer[i] = 0;
         append(key, buffer);
         i = 0;
      }
   }

   fclose(fin);

   return 1;
}

void Spider::QueryPage(char *filename, char *name, int page, char *key)
{
   char lnkfile[256];


   sprintf(lnkfile, "%s.links", name);

   parse_links(filename, lnkfile, 4);

   /* Now you should have a LINKS file full of links to more
      links for your original query. */

   QueryNextPage(lnkfile, name, page, key);

   /* Remove cache files, is this safe? */
   unlink(lnkfile);
}

void Spider::Handle(int err, char *url, char *filename)
{
   switch(err)
   {
   case 1:
      printf("Adding to success.cache.\n");
      append("success.cache", url);
      break;
   case 421:
      printf("Adding to busy.cache.\n");
      append("busy.cache", url);
      break;
   case 550:
       printf("Adding to crack.cache.\n");
      //remove_url("busy.cache", url); // FIXME
      append("crack.cache", url);
      break;
   }
}


int Spider::Query(char *query, int type)
{
   char *url;
   char *key;
   char c;
   int j = 0, i = 0, err;
   FILE *f;


   url = new char[strlen(query) + 128];

   while((c = query[i++]) != 0)
   {
      switch (c)
      {
      case ' ':
         query[i - 1] = '+';
         break;
      default:
         ;
      }
   }
      
   switch(type)
   {
   case LYCOS_MP3:
      for (i = 0, j = strlen(query); i < j; i++)
	 if (query[i] == ' ')
	    query[i] = '+';

      sprintf(url, "%s%s%s", 
              "http://mp3search.lycos.com/cgi-bin/search?form=lycosnet&query=",
              query,
              "&doit=Go+Get+It%21");

      for (i = 0, j = strlen(query); i < j; i++)
	 if (query[i] == '+')
	    query[i] = ' ';      
      break;
   default:
      printf("Search type %i isn't vaild.\n", type);
      return 0;
   }

   j = strlen(url);

   err = Get(url, "spider.tmp");

   f = fopen("spider.tmp", "r");

   if (f)
      fclose(f);
   else
   {
      printf("\n   ERROR: url could not be retrieved.\n");
      printf("   Aborting query.\n\n");

      delete url;
      shutdown_goosehunt();
   }

   key = clean_filename(query);

   QueryPage("spider.tmp", "spider", 0, key);

   delete url;
   /* Remove cache files, is this safe? */
   unlink("spider.tmp");

   return err;
}

int Spider::Get(char *url_string, char *filename)
{
   int retval = -1;
   Url *url = new Url(url_string);
   Ftp *ftp = new Ftp();
   Http *http = new Http();
   Gopher *gopher = new Gopher();


   if (!url_string || strcmp(url_string, "") == 0)
      return -2;


   /* If filename exists should we append numeric to string? */
   //cout << endl << endl << "Spider->get(\"" << url_string << "\") " << endl;

   if (PROXY_URL)
   {
      url->Proxy(PROXY_URL, HTTP_PROXY | FTP_PROXY | GOPHER_PROXY);
      printf("==Using Proxy Server==\n");
   }

   url->Options(OPT_PROGRESS | OPT_RESUME | OPT_BE_MOZILLA | OPT_VERBOSE);

   url->Print();

   switch(url->Protocol())
   {
   case FTP_SESSION:
      retval = ftp->Fetch(url, filename);
      break;
   case HTTP_SESSION:
      retval = http->Fetch(url, filename);
      break;
   case GOPHER_SESSION:
      retval = gopher->Fetch(url, filename);
      break;
   default:
      printf("Sorry unknown protocol in url.\n");
   }

   delete url;
   delete ftp;
   delete http;
   delete gopher;

   printf("***Fetch status [%s] = ", filename);

   switch(retval)
   {
   case 1:
      printf("successful\n");
      break;
   case 404:
      printf("not found\n");
      break;
   case 421:
      printf("busy\n");
      break;
   case 530:
      printf("bad login\n");
      break;
   case 550:
   case 553:
      printf("denied\n");
      break;
   default:
      printf("unknown error %i\n", retval);
      printf("Email stu7440@westga.edu about adding this error code.\n");
   }

   printf("\n-------------------------------------------------\n\n");

   return retval;
}

void Spider::Fetch(char *key)
{
   FILE *f;
   char c;
   char url[256];
   char filename[256];
   int count = 0, err = 0, t = 0, percent = 0;
   int busy = 0, denied = 0, bad_login = 0;
   int url_count = 0, good = 0, unknown = 0;
   int not_found = 0;


   count = 0;

   f = fopen(key, "r");

   if (!f)
   {
      strcpy(key, clean_filename(key));

      f = fopen(key, "r");

      if (!f)
      {
  	 printf("File %s not found.\n", key);
         return;
      }
   }

   while(fscanf(f, "%c", &c) != EOF)
   {
      if (count > 254)
      {
	 /* Handle this later, not a big deal since most urls < 128 */
 	 printf("Spider: URL in fetch cache too long, aborting.\n");
         return;
      }

      url[count++] = c;

      if (c == '/')
         t = 0;
      else
         filename[t++] = c;

      if (c == '\n')
      {
	 /* Comment not url */
         if (url[0] == '#')
	 {
            count = 0;
            continue;
         }

         url[count - 1] = 0;
         filename[t - 1] = 0;

         strcpy(filename, clean_filename(filename));

         err = Get(url, filename);
         Handle(err, url, filename);

         url_count++;

         switch(err)
	 {
         case -2:
            url_count--;
	    break;
         case 1:
            good++;
            break;
         case 404:
            not_found++;
            break;
         case 421:
            busy++;
            break;
         case 530:
            bad_login++;
            break;
         case 550:
         case 553:
            denied++;
            break;
         default:
            unknown++;
	 }

         count = 0;
      }
   }

   if (url_count == 0)
   {
      printf("No urls to fetch\n");
      fclose(f);
      return;
   }

   printf("\nFetch on %i url(s)\n", url_count);
   
   if (bad_login)
   {
      percent = (int)(100.0 * (double) bad_login / url_count);
      printf(" bad login: %i\t%i%c\n", bad_login, percent, '%');
   }

   if (denied)
   {
      percent = (int)(100.0 * (double) denied / url_count); 
      printf(" denied:    %i\t%i%c\n", denied, percent, '%');
   }

   if (busy)
   {
      percent = (int)(100.0 * (double) busy / url_count); 
      printf(" busy:      %i\t%i%c\n", busy, percent, '%');
   }

   if (not_found)
   {
      percent = (int)(100.0 * (double) not_found / url_count);
      printf(" not found: %i\t%i%c\n", not_found, percent, '%');
   }

   percent = (int)(100.0 * (double) unknown / url_count);
   printf(" unknown:   %i\t%i%c\n", unknown, percent, '%');

   percent = (int)(100.0 * (double) good / url_count);
   printf(" fetched:   %i\t%i%c\n", good, percent, '%');

   fclose(f);
}

